//
//  Guit GUI Toolkit
//  Copyright © 2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <guit.hpp>
#include "examples.hpp"
using namespace std;
using namespace guit;

class BoxesAndBars : public GWindow {
public:
  BoxesAndBars() : GWindow("#boxes-bars Boxes and Bars") {
    auto& choice = Choice();
    auto& status = Label("Ready");

    auto action = [&](auto& e) {
      if (!e.chosen()) status = "";
      else status = "Item #" + toString(e.chosen()+1) + ": " + e.chosen()->text();
    };

    // - - - Boxes - - - - - - - - - - - - -

    GStrings items{"Apples", "Oranges", "Mangoes"};

    auto& boxes = HBox("#row1")
    << (Box("Box")             << items)
    << (ButtonBox("ButtonBox") << items)
    << (RadioBox("RadioBox")   << items << choice << action)
    << (ChoiceBox("ChoiceBox") << items << choice)
    << (ListBox("ListBox")     << items << choice)
    << (OptionBox("OptionBox") << items << choice)
    << (ComboBox("ComboBox")   << items << choice)
    << (TabBox("TabBox")       << items << choice);

    // - - - Charts and Tabs - - - - - - - - - - - - - -

    GStrings valitems {
      "{background: yellow; value: 30} Apples",
      "{background: orange; value: 60} Oranges",
      "{background: red; value: 40;} Mangoes"
    };

    auto& charts = HBox("#charts")
    << (ChartBox(".chart {flow:row} Row")       << valitems << choice)
    << (ChartBox(".chart {flow:column} Column") << valitems << choice)
    << (ChartBox(".chart {flow:pie} Pie")       << valitems << choice);

    auto& tab = TabBox("#tabbox Boxes in TabBox")
    << choice
    << (Box(".tabchild Apples")  << "I like Apples")
    << (Box(".tabchild Oranges") << "I like Oranges")
    << (Box(".tabchild Mangoes") << "I like Mangoes");

    auto& charts_and_tabs = HBox("#row2")
    << charts
    << tab;

    // - - - Menus - - - - - - - - - - - - -

    auto& file_menu = Menu("File")
    << "{icon: open.png} Open"
    << "{icon: save.png} Save"
    << Separator
    << ("{icon: quit.png} Quit" << []{ GApp::close(); });

    auto& fruit_menu = Menu("Fruits") << items << choice;

    // - - - Window - - - - - - - - - - - - -

    *this
    << (MenuBar() << file_menu << fruit_menu)
    << (ToolBar() << choice << items)
    << (StatusBar() << status)
    << boxes
    << charts_and_tabs;
  }
};


GWindow& Examples::makeBoxesAndBars() {
  return *new BoxesAndBars();
}
