//
//  Socket client example
//  Guit GUI Toolkit
//  Copyright © 2019/2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <guit.hpp>
#include <gsocket.hpp>
#include "examples.hpp"
using namespace std;
using namespace guit;


class RemoteApp : public GWindow {
  bool server_mode{};
  GSocket sock;
  GText hostname{"localhost"};
  GInt port{3331};
  GCheck running;
  GChoice tab{0}, light{0}, channel{0};
  GRegText from{"00:00"}, to{"00:00"};

public:
  RemoteApp(bool server_mode) :
  GWindow("#remote-app Remote App"),
  server_mode(server_mode)
  {
    // Props bounds with those of the remote app
    sock["tab"]     ^= tab;
    sock["light"]   ^= light;
    sock["channel"] ^= channel;
    sock["from"]    ^= from;
    sock["to"]      ^= to;
    sock["running"] ^= running;

    from.setTimeFormat();
    to.setTimeFormat();
    
    auto& tabs = TabBox("{border:rounded;}")
    << tab
    << (ListBox("#lights Lights")     << light)
    << (ListBox("#channels Channels") << channel);

    auto& time_box = Pane("<b>Time")
    << (Table
        << (TRow() << "From:" << from)
        << (TRow() << "To:"   << to)
        << (Button()
            << running
            << on::checked   / "Stop"  / [this]{run(true);}
            << on::unchecked / "Start" / [this]{run(false);}
            )
        );
    
    auto& connect_box = Pane("{size:15c} <b>Connection")
    << (Table()
        << (TRow()
            << "Status:"
            << ("      "
                << (~sock.status() == 1) / GBackground::green
                << (~sock.status() != 1)  / GBackground::orange
                )
            )
        << (TRow() << "Host:" << hostname)
        << (TRow() << "Port:" << port)
        << (Button()
            << (~sock.status() != 1) / "Connect"
            << (~sock.status() == 1) / "Disconnect"
            << [this]{connectDisconnect();}
            )
        );
      
        
    GString title;
    if (!server_mode) title = "Remote - Client";
    else {
      title = "Remote - Server";
      hostname = "-- server --";
    }
    
    *this
    << (HBox("{align: center center}")
        << (Button(" Create Client ") << [this]{createClient();})
        )
    << (HBox()
        << (VBox("#leftbox")  << tabs)
        << (VBox("#rightbox") << time_box << connect_box)
        );
  }

  void connectDisconnect() {
    if (sock.status() != 1) {     // connect
      if (server_mode) sock.startAsServer(port);
      else sock.startAsClient(hostname(), port);
    }
    else sock.disconnect();
  }
  
  void run(bool state) {
    cout << "Running: " <<  state << endl;
  }

  void createClient() {
    system((GApp::argv(0) + " -client &"));
  }
};


GWindow& Examples::makeServerApp() {
  return *new RemoteApp(true);
}

GWindow& Examples::makeClientApp() {
  return *new RemoteApp(false);
}
