//
//  Gadget Styles
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
#ifndef Guit_GadgetStyle_hpp
#define Guit_GadgetStyle_hpp
#include <gprop.hpp>
namespace guit {

class GStyle;

/// Styles of a gadget instance.
/// List of GStyles corresponding to the class and id names given to the constr.
class GadgetStyle : public GObject {
public:
  ~GadgetStyle();

  static bool initStyle(Gadget&, GString const& spec, GString& title);
  void realize(Gadget const&);

  GProp* findProp(Gadget const&, GPropType&) ;

  template <class T> T* findProp(Gadget const& g)  {
    return static_cast<T*>(findProp(g, T::Type()));
  }

  // either the 1st or 2nd argument must be non-nul.
  static bool parseProps(Gadget*, GPropList*, GString const& style,
                         GString const& delim);

  GString typeName() const override;

private:
  friend class GStyle;
  friend class GRender;
  GadgetStyle(GString const& ids);
  bool realized_{};
  uint16_t size_{};
  union {
    GString* names_;
    GId* ids_;  // null or 0 ended
  };
};

}
#endif
