
//
//  Gadget IO
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef guit_iobuffer_hpp
#define guit_iobuffer_hpp

#include <iostream>
#include <gstring.hpp>
namespace guit {

class GIOBuffer {
public:
  struct IOError {};

  GIOBuffer(std::ostream&);
  GIOBuffer(std::istream&);

  GString const& data() const {return data_;}
  bool valueMode() const {return valuemode_;}
  void setValueMode(bool);

  bool read();
  void clear();
  void printTabbed(GString const&);
  void printDelim();

  int level() const {return level_;}
  void incrLevel() {++level_;}
  void decrLevel() {--level_;}
  void resetLevel() {level_ = 0;}

  int tabCount() const {return tabcount_;}
  void setInsideRow(bool);

private:
  friend class Gadget;
  bool begin_{true}, valuemode_{}, skipbase_{}, insiderow_{};
  int  level_{}, tabcount_{}, hasdata_{};
  std::ostream* out_{};
  std::istream* in_{};
  GString data_;
};

}
#endif

