//
//  Viewport for scrolling another gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gtrigger.hpp>
#include <gevent.hpp>
#include <gscrollbox.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <grender.hpp>
#include <gstyle.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(ScrollBox)

//            TODO:  UN SEUL CHILD !!!!

GScrollBox::~GScrollBox() = default;

GScrollBox::GScrollBox(GadgetType&t, GString const& spec) : GBox(t, spec),
hscrollbar_(new GScrollBar),
vscrollbar_(new GScrollBar)
{
  eventmask_ = eventmask_ | GEventType::Wheel;
  *hscrollbar_ << GFlow::row;
  *vscrollbar_ << GFlow::column;
  vscrollbar_->modes_.internal = 1;
  hscrollbar_->modes_.internal = 1;
  borderAdd(hscrollbar_, Place::Bottom);
  borderAdd(vscrollbar_, Place::Right);
}

GScrollBox::GScrollBox(GString const& spec)
: GScrollBox(Type(), spec) {}

GScrollBox& GScrollBox::clone(GClone const& cm) {
  auto* g = new GScrollBox;
  g->initClone(*this, cm);
  return *g;
}

GScrollBar& GScrollBox::hscrollbar() {return *hscrollbar_;}
GScrollBar& GScrollBox::vscrollbar() {return *vscrollbar_;}

float GScrollBox::hscroll() const {return hscrollbar_->value();}
float GScrollBox::vscroll() const {return vscrollbar_->value();}

void GScrollBox::setScroll(float xscroll, float yscroll) {
  if (xscroll < 0.f) xscroll = 0.f;
  else if (xscroll > 1.f) xscroll = 1.f;

  if (yscroll < 0.f) yscroll = 0.f;
  else if (yscroll > 1.f) yscroll = 1.f;

  GRender r(*this, nullptr);   // le parent_render est null !!!!
  scrollImpl(r, xscroll, yscroll);
  hscrollbar_->setValue(xscroll);
  vscrollbar_->setValue(yscroll);
}

void GScrollBox::setScrollDelta(float dx, float dy) {
  setScroll(hscrollbar_->value() + dx, vscrollbar_->value() + dy);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GScrollBox::pack(GRender* parr) {
  GRender r(*this, parr);
  scrollImpl(r, hscrollbar().value(), vscrollbar().value());
}

GRect GScrollBox::scrollImpl(GRender& r, float xvalue, float yvalue) {
  GRect ch_area = childrenArea(r);
  packBorders(ch_area, r);  // changes ch_area!

  Gadget* ch = childAt(0);    // UN SEUL CHILD !!!!
  if (!ch) {
    //xvalue = 0.f;
    //yvalue = 0.f;
    return ch_area;
  }
  GRect chframe_init = ch->frame_;

  //auto& rowflex = ch->rowFlexRef(this);
  auto rowflex = ch->rowFlex(this);
  // Fixed: nothing to do, keeps size
  // Other value: grows, except for MultiLine text
  if (rowflex != 0) {
    // Grow: takes the max of previous size and new size
    if (!ch->modes_.multiLine) {
      //rowflex = -1; // means growOnly, used in GRender::computeSize()
      ch->setRowFlex(this, -1);
      ch->frame_.w = std::max(ch_area.w, ch->frame_.w);
    }
    else { // multiline text: adapts
      ch->frame_.w = ch_area.w;
      // increase size if the text goes not reach the bottom of the ch_area
      if (ch->frame_.h < ch_area.h) ch->frame_.h = ch_area.h;
    }
  }

  //auto& colflex = ch->columnFlexRef(this);
  auto colflex = ch->columnFlex(this);
  // Fixed: nothing to do, keeps size
  // Other value: grows, except for MultiLine text
  if (colflex != 0) {
    if (!ch->modes_.multiLine) {
      //colflex = -1;  // means growOnly
      ch->setColumnFlex(this, -1);
      ch->frame_.h = std::max(ch_area.h, ch->frame_.h);
    }
    else {
      // MultiLine text adapts
      // the vertical size must be recomputed because horizontal size changed.
      // false arg means that the Size prop is not taken into account (if any)
      //GRender(*child, &r).computeSize(false);
      ch->computeSize(&r, false);
      // increase size if the text goes not reach the bottom of the ch_area
      if (ch->frame_.h < ch_area.h) ch->frame_.h = ch_area.h;
    }
  }

  ch->pack(&r);

  // force MultiLine text to adapt
  if (rowflex > 0 && ch->modes_.multiLine) {
    if (ch->frame_.w < ch_area.w) ch->frame_.w = ch_area.w;
    if (ch->frame_.h < ch_area.h) ch->frame_.h = ch_area.h;
  }
  ch->movedResizedCB(chframe_init);
  
  float chw = ch->frame_.w, chh = ch->frame_.h;
  float xoffset = 0.f;
  xoffset = (ch_area.w - chw) * xvalue;
  if (xoffset >= 0.f) xoffset = xvalue = 0.f;
  
  float yoffset = 0.f;
  yoffset = (ch_area.h - chh) * yvalue;
  if (yoffset >= 0.f) yoffset = yvalue = 0.f;

  // nb: calls Move callback
  ch->moveImpl(xoffset + ch_area.x, yoffset + ch_area.y, true);
  update();
  return ch_area;
}

}
