//
// Graphics Layer
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#if defined(_WIN32)
#  include <windows.h>
#  undef min
#  undef max
#endif
#if defined(__APPLE__)
#  define GL_SILENCE_DEPRECATION
#  include <OpenGL/gl.h>
#else
#  include <GL/gl.h>
#endif

#include <iostream>
#include <cmath>
#include <gapp.hpp>
#include <ggraphics.hpp>
#include <gevent.hpp>
#include <gcolor.hpp>
#include <gfont.hpp>
#include <gicon.hpp>
#include <gadget.hpp>
#include <gwindow.hpp>
#include <gcontrol.hpp>
#include <gnative.hpp>
using namespace std;
namespace guit {

GGraphics::GGraphics(GEvent& e) :
gly_(0.f),
thickness_(1.f),
color_(0.f, 0.f, 0.f, 1.f),
bgcolor_(1.f,1.f,1.f,1.f)
{
  win_ = e.window();
  if (auto* g = e.gadget_) {
    frame_ = g->wframe();
    gly_ = win_->frame_.h - frame_.y - frame_.h;
    //setClip(frame); inutile et faux!
  }
  else gerror("GGraphics", "Invalid event (no sender)");
}

GGraphics::GGraphics(GWindow& win) :
gly_(0.f),
thickness_(1.f),
color_(0.f, 0.f, 0.f, 1.f),
bgcolor_(1.f, 1.f, 1.f, 1.f),
win_(&win) {}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GGraphics::makeCurrent() {
  win_->natwin_->makeCurrent();
}

void GGraphics::swapBuffers() {
  win_->natwin_->swapBuffers();
}

/*
 void GGraphics::makeCurrent(LGLContext* c) {
 win_->natwin()->makeCurrent(c);
 }

 GLContext* GGraphics::createGLContext() {
 return win_->natwin()->createGLContext();
 }
 */

void GGraphics::setScale(float scale) {
  scale_ = scale;
}

void GGraphics::setWinClip(const GRect& clip) {
  win_->natwin_->setClip(clip.x, clip.y, clip.w, clip.h);
}

void GGraphics::setWinClip(float x, float y, float w, float h) {
  win_->natwin_->setClip(x, y, w, h);
}

void GGraphics::save() {
  savex_ = frame_.x;
  savey_ = frame_.y;
}

void GGraphics::restore() {
  frame_.x = savex_;
  frame_.y = savey_;
}

void GGraphics::translate(float x, float y) {
  frame_.x += x;
  frame_.y += y;
}

void GGraphics::setThickness(float thickness) {
  thickness_ = thickness;
}

void GGraphics::unsetColor() {
  drawfg_ = false;
}

void GGraphics::setColor(const GRgba* c) {
  drawfg_ = (c && c->alpha > 0.f);
  if (c) color_ = *c;
}

void GGraphics::setColor(const GColor& c) {
  setColor(&c.value());
}

void GGraphics::setColor(const GColor* c) {
  setColor(c ? &c->value() : nullptr);
}

void GGraphics::setColor(GColor const& color, float a) {
  GRgba c = color();
  c.alpha = a;
  setColor(&c);
}

void GGraphics::unsetBgColor() {
  drawbg_ = false;
}

void GGraphics::setBgColor(const GRgba* c) {
  drawbg_ = (c && c->alpha > 0.f);
  if (c) bgcolor_ = *c;
}

void GGraphics::setBgColor(const GColor& c) {
  setBgColor(&c.value());
}

void GGraphics::setBgColor(const GColor* c) {
  setBgColor(c ? &c->value() : nullptr);
}

void GGraphics::setBgColor(GColor const& color, float a) {
  GRgba c = color();
  c.alpha = a;
  setBgColor(&c);
}

void GGraphics::setFont(GFont const& font) {
  font_ = font.value();
}

void GGraphics::setFont(GFontValue const& font) {
  font_ = font;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*
class GTextDecoration {
public:
  bool overline{}, underline{}, linethrough{}, frame{};
  const GRgba *color{}, *bgcolor{};
};*/

/// y = lower point, should be baseline !!
GDim GGraphics::drawString(float x, float y, GString const& str
                           /*, GTextDecoration const* dec*/) {
  if (!font_.realize()) return GDim();

  //if (str.empty()) return font_.textSize(str, false, nullptr);

  //GDim textsize = font_.textSize(str, false, nullptr);
  GDim textsize = GNatContext::textSize(*font_.impl, str);
  if (str.empty()) return textsize;

  /*
  if (dec && (dec->bgcolor || dec->color)) {
    auto bg = bgcolor_;
    auto fg = color_;
    auto drawbg = drawbg_;
    auto drawfg = drawfg_;

    if (dec->bgcolor) {
      drawbg_ = true;
      bgcolor_ = *dec->bgcolor;
    }
    else drawbg_ = false;

    if (dec->frame && dec->color) {
      drawfg_ = true;
      color_ = *dec->color;
    }
    else drawfg_ = false;

    drawRect(GRect(x, y-textsize.h, textsize.w, textsize.h));

    if (dec->color) {
      color_ = *dec->color;
      if (dec->underline) drawLine(x, y, x+textsize.w, y);
      if (dec->overline) drawLine(x, y-textsize.h, x+textsize.w, y-textsize.h);
      if (dec->linethrough) drawLine(x, y-textsize.h/2.f, x+textsize.w, y-textsize.h/2.f);
    }
    bgcolor_ = bg;
    color_ = fg;
    drawbg_ = drawbg;
    drawfg_ = drawfg;
  }*/

  if (bgcolor_.alpha > 0.f) {
    bool drawbg = drawbg_, drawfg = drawfg_;
    drawbg_ = true;
    drawfg_ = false;
    drawRect(x, y-textsize.h, textsize.w, textsize.h);
    drawbg_ = drawbg;
    drawfg_ = drawfg;
  }

  GTexture t = GNatContext::createFontTexture(str, *font_.impl);
  if (!t) {
    gerror("drawString","Could not create font texture for: '" + str +"'");
    return textsize;
  }
  drawTexture(x, y-textsize.h, textsize.w, textsize.h, t, &color_);
  deleteTexture(t);
  return textsize;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GGraphics::drawLine(float x1, float y1, float x2, float y2) {
  beginGL();
  glLineWidth(thickness_);  // muste be before glBegin() !
  glBegin(GL_LINE_STRIP);
  glColor4f(color_.red, color_.green, color_.blue, color_.alpha);
  glVertex2f(frame_.x+x1, frame_.y+y1);
  glVertex2f(frame_.x+x2, frame_.y+y2);
  glEnd();
  endGL();
}

void GGraphics::drawPolyline(const std::vector<GPoint>& points) {
  beginGL();
  glLineWidth(thickness_);
  glBegin(GL_LINE_STRIP);
  glColor4f(color_.red, color_.green, color_.blue, color_.alpha);
  for (auto& p : points) glVertex2f(frame_.x+p.x, frame_.y+p.y);
  glEnd();
  endGL();
}

void GGraphics::drawPolygon(const std::vector<GPoint>& points) {
  beginGL();
  if (drawbg_) {
    glBegin(GL_POLYGON);
    glColor4f(bgcolor_.red, bgcolor_.green, bgcolor_.blue, bgcolor_.alpha);
    for (auto& p : points) glVertex2f(frame_.x+p.x, frame_.y+p.y);
    glEnd();
  }
  if (drawfg_) {
    glLineWidth(thickness_);
    glBegin(GL_LINE_LOOP);
    glColor4f(color_.red, color_.green, color_.blue, color_.alpha);
    for (auto& p : points) glVertex2f(frame_.x+p.x, frame_.y+p.y);
    glEnd();
  }
  endGL();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GGraphics::drawRect(float x, float y, float w, float h) {
  if (h <= 0 || w <= 0) return;
  x += frame_.x;
  y += frame_.y;
  /*
   GLfloat w1[2]{}, w2[2]{};
   glGetFloatv(GL_ALIASED_LINE_WIDTH_RANGE, w1);
   glGetFloatv(GL_SMOOTH_LINE_WIDTH_RANGE, w2);
   cerr << "thickness_ " << thickness_ << " " << w1[1] << " " << w2[1] << endl;
   */
  beginGL();
  if (drawbg_) {
    glColor4f(bgcolor_.red, bgcolor_.green, bgcolor_.blue, bgcolor_.alpha);
    glRectd(x, y, x+w, y+h);
  }
  if (drawfg_) {
    glLineWidth(thickness_);   // msut be before glBegin() !
    glBegin(GL_LINE_LOOP);
    glColor4f(color_.red, color_.green, color_.blue, color_.alpha);
    glVertex2d(x, y);
    glVertex2d(x+w, y);
    glVertex2d(x+w, y+h);
    glVertex2d(x, y+h);
    glEnd();
  }
  endGL();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static void drawRectArc(float center_x, float center_y, float aw, float ah,
                        float from, float to, double incr) {
  for (double a = from; a >= to; a -= incr) {
    glVertex2d(center_x + aw * std::cos(a), center_y - ah * std::sin(a));
  }
}

void GGraphics::drawRoundRect(float x, float y, float w, float h,
                             float arc_w, float arc_h) {
  if (h <= 0 || w <= 0) return;
  x += frame_.x;
  y += frame_.y;
  float x2 = x + w;
  float y2 = y + h;
  float aw = arc_w, ah = arc_h; // arc radius
  float incr = float(M_PI / std::max(aw,ah));  // resolution
  
  beginGL();
  if (drawbg_) {
    glLineWidth(thickness_);
    glColor4f(bgcolor_.red, bgcolor_.green, bgcolor_.blue, bgcolor_.alpha);
    //glEnable(GL_POLYGON_SMOOTH);
    //glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    glBegin(GL_POLYGON);
    
    glVertex2d(x+aw, y); glVertex2d(x2-aw, y); //top line
    drawRectArc(x2-aw, y+ah, aw, ah, float(M_PI/2), 0, incr); // top right arc
    
    glVertex2d(x2, y+ah); glVertex2d(x2, y2-ah); //right line
    drawRectArc(x2-aw, y2-ah, aw, ah, 0, float(-M_PI/2), incr); // bottom right arc
    
    glVertex2d(x2-aw, y2); glVertex2d(x+aw, y2); //bottom line
    drawRectArc(x+aw, y2-ah, aw, ah, float(-M_PI/2), float(-M_PI), incr); // bottom left arc
    
    glVertex2d(x, y2-ah); glVertex2d(x, y+ah); //left line
    drawRectArc(x+aw, y+ah, aw, ah, float(M_PI), float(M_PI/2), incr); // top left arc
    
    glEnd();
    //glDisable(GL_POLYGON_SMOOTH);
  }
  
  if (drawfg_) {
    glLineWidth(thickness_);
    glColor4f(color_.red, color_.green, color_.blue, color_.alpha);
    //pose pbm avec fonts!
    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    //glBegin(GL_POLYGON);
    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glBegin(GL_LINE_LOOP);
    
    glVertex2d(x+aw, y); glVertex2d(x2-aw, y); //top line
    drawRectArc(x2-aw, y+ah, aw, ah, float(M_PI/2), 0, incr); // top right arc
    
    glVertex2d(x2, y+ah); glVertex2d(x2, y2-ah); //right line
    drawRectArc(x2-aw, y2-ah, aw, ah, 0, float(-M_PI/2), incr); // bottom right arc
    
    glVertex2d(x2-aw, y2); glVertex2d(x+aw, y2); //bottom line
    drawRectArc(x+aw, y2-ah, aw, ah, float(-M_PI/2), float(-M_PI), incr); // bottom left arc
    
    glVertex2d(x, y2-ah); glVertex2d(x, y+ah); //left line
    drawRectArc(x+aw, y+ah, aw, ah, float(M_PI), float(M_PI/2), incr); // top left arc
    
    glEnd();
    glDisable(GL_LINE_SMOOTH);
  }
  endGL();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool GGraphics::pickOval(GRect const& r, GPoint const& pt) {
  auto w = r.w/2;
  auto h = r.h/2;
  auto cx = r.x + w;
  auto cy = r.y + h;
  return ((pt.x-cx) / w) * ((pt.x-cx) / w) + ((pt.y-cy) / h) * ((pt.y-cy) / h) <= 1.;
}

bool GGraphics::pickPie(GRect const& r, float start, float ext, GPoint const& pt) {
  if (r.w < 0.f || r.h < 0.f) return false;
  auto w = r.w/2;
  auto h = r.h/2;
  auto cx = r.x + w;
  auto cy = r.y + h;
  auto dx = pt.x - cx;
  auto dy = cy - pt.y;
  if (((pt.x-cx) / w) * ((pt.x-cx) / w) + ((pt.y-cy) / h) * ((pt.y-cy) / h) > 1.)
    return false;
  
  auto angle = atan(dy / dx);
  if (dy >= 0) {
    if (dx < 0) angle += float(M_PI);
  }
  else {
    if (dx < 0) angle += float(M_PI);
    else angle += float(2 * M_PI);
  }
  auto astart = start * M_PI / 180.;
  auto aend = astart + ext * M_PI / 180.;
  return angle >= astart && angle <= aend;
}

void GGraphics::drawArc(float x, float y, float w, float h,
                       float start, float ext, bool drawpie) {
  if (w < 0.f || h < 0.f) return;
  x += frame_.x;
  y += frame_.y;
  float rw = w / 2.f;  // width radius
  float rh = h / 2.f;  // height radius
  float cx = x + rw;
  float cy = y + rh;
  float incr = float(M_PI / max(w, h));  // resolution
  
  double a1 = -start * M_PI / 180.;   // counterclockwise => -start
  double a2{}, a3{};
  if (ext >= 360.f) ext = 361.f;  // to avoid discontinuties
  else if (ext <= -360.f) ext = -361.f;
  
  bool concave = false;
  if (ext > 180.f && ext < 360.f) {
    a2 = (-start-180) * M_PI / 180.;
    concave = true;
  }
  else if (ext < -180.f && ext > -360.f) {
    a2 = (-start+180) * M_PI / 180.;
    concave = true;
  }
  else {
    a2 = (-start-ext) * M_PI / 180.;
  }
  a3 = (-start-ext) * M_PI / 180.;
  //cerr << a1 <<" " << a2 << " "<< endl;
  
  beginGL();
  
  if (drawpie && drawbg_) {
    glColor4f(bgcolor_.red, bgcolor_.green, bgcolor_.blue, bgcolor_.alpha);
    
    glEnable(GL_POLYGON_SMOOTH);   // marche pas ?
    //glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    
    glBegin(GL_POLYGON);
    if (ext < 0) {
      for (double a = a1; a <= a2; a += incr)
        glVertex2d(cx + rw * std::cos(a), cy + rh * std::sin(a));
    }
    else {
      for (double a = a1; a >= a2; a -= incr)
        glVertex2d(cx + rw * std::cos(a), cy + rh * std::sin(a));
    }
    // add central point to get a quadrant
    if (::fabs(ext) < 360) glVertex2d(cx, cy);
    glEnd();
    
    if (concave) {
      glBegin(GL_POLYGON);
      if (ext < 0) {
        for (double a = a2-incr; a <= a3; a += incr)
          glVertex2d(cx + rw * std::cos(a), cy + rh * std::sin(a));
      }
      else {
        for (double a = a2+incr; a >= a3; a -= incr)
          glVertex2d(cx + rw * std::cos(a), cy + rh * std::sin(a));
      }
      glVertex2d(cx, cy);
      glEnd();
    }
    glDisable(GL_POLYGON_SMOOTH);
  }
  
  if (drawfg_) {
    glLineWidth(thickness_);
    glColor4f(color_.red, color_.green, color_.blue, color_.alpha);
    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

    glBegin(GL_LINE_STRIP);

    if (drawpie && ::fabs(ext) < 360) glVertex2d(cx, cy);

    if (ext < 0) {
      for (double a = a1; a <= a3; a += incr)
        glVertex2d(cx + rw * std::cos(a), cy + rh * std::sin(a));
    }
    else {
      for (double a = a1; a >= a3; a -= incr)
        glVertex2d(cx + rw * std::cos(a), cy + rh * std::sin(a));
    }

    if (drawpie && ::fabs(ext) < 360) glVertex2d(cx, cy);

    glEnd();
    glDisable(GL_LINE_SMOOTH);
  }
  
  endGL();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool GGraphics::readImage(GIcon& ima) {
  return readImage(ima.value_);
}

bool GGraphics::readImage(GIconValue& ima) {
  if (ima.status == GIconValue::Loaded) return true;
  ima.texture = 0;
  ima.status = GIconValue::Unloaded;
  
  if (ima.drawfun == GIcon::Image) {
    GNatContext::readTexture(ima.filename, ima);
  }
  else if (ima.drawfun == GIcon::XPM) {
    GNatContext::readTexture(ima.xpm, ima);
  }
  
  if (ima.status == GIconValue::Loaded) return true;
  else {
    ima.status = GIconValue::CantLoad;
    gerror("GGraphics::loadImage", "Can't load Image '"+ima.filename+"'");
    return false;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GGraphics::drawImage(GRect const& rect, GIcon& ima) {
  drawImage(rect, ima.value_);
}

void GGraphics::drawImage(float x, float y, float w, float h, GIconValue& ima) {
  if (ima.status != GIconValue::Loaded) {
    if (ima.status == GIconValue::CantLoad) return;
    else if (!readImage(ima)) return;
  }
  drawTexture(x,y,w,h, ima.texture, &GColor::white.value_);
  // n'affiche pas les images si le fond est noir !
  //drawTexture(x,y,w,h, ima.texture);
}

void GGraphics::drawTexture(float x, float y, float w, float h,
                           GTexture texture, GRgba const* color) {
  beginGL();
  GLfloat
  x1 = frame_.x + x,
  y1 = frame_.y + y,
  x2 = x1 + w,
  y2 = y1 + h;
  
  glPushAttrib(GL_CURRENT_BIT);
  glEnable(GL_TEXTURE_2D);
  
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glEnable(GL_BLEND);
  //glColor4f(1.0, 1.0, 1.0, g.alpha);
  
  if (color) glColor4f(color->red, color->green, color->blue, color->alpha);

  //else glColor4f(1.0, 1.0, 1.0, 1.);
  //else glColor4f(color_.red, color_.green, color_.blue, color_.alpha);
  
  glBindTexture(GL_TEXTURE_2D, texture);
  glBegin(GL_QUADS);
  glTexCoord2f(0.0, 0.0); glVertex2f(x1, y1);
  glTexCoord2f(1.0, 0.0); glVertex2f(x2, y1);
  glTexCoord2f(1.0, 1.0); glVertex2f(x2, y2);
  glTexCoord2f(0.0, 1.0); glVertex2f(x1, y2);
  glEnd();
  
  //if (alpha >= 1.) glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);
  glPopAttrib();
  endGL();
}

void GGraphics::deleteTexture(GLuint texture) {
  glEnable(GL_TEXTURE_2D);
  glDeleteTextures(1, &texture);
  glDisable(GL_TEXTURE_2D);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// A big number, but not so big it can't survive a few float operations.
static const float BIG_VALUE = FLT_MAX / 10.0;

static bool linesIntersect(const GPoint& a, const GPoint& b,
                           const GPoint& c, const GPoint& d);
static int evaluateCrossings(GPoints const& points,
                             float x, float y, bool useYaxis, float distance);

GClip::GClip(const GRect& a, const GRect& b) {
  x1 = std::max(a.x, b.x);
  y1 = std::max(a.y, b.y);
  x2 = std::min(a.x + a.w, b.x + b.w);
  y2 = std::min(a.y + a.h, b.y + b.h);
}

void GClip::intersection(const GRect& b) {
  x1 = std::max(x1, b.x);
  y1 = std::max(y1, b.y);
  x2 = std::min(x2, b.x + b.w);
  y2 = std::min(y2, b.y + b.h);
}

bool GClip::contain(const GPoint& point) const {
  return point.x >= x1 && point.x <= x2 && point.y >= y1 && point.y <= y2;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GPoint GPoint::operator+(GPoint const& p) const {
  return GPoint(x + p.x, y + p.y);
}

GPoint GPoint::operator-(GPoint const& p) const {
  return GPoint(x - p.x, y - p.y);
}

bool GPoint::operator==(const GPoint& other) const {
  return x == other.x && y == other.y;
}

bool GPoint::operator!=(const GPoint& other) const {
  return x != other.x || y != other.y;
}

std::ostream& operator<<(std::ostream& out, const GPoint& p) {
  out << p.x << " "<< p.y;
  return out;
}

std::istream& operator>>(std::istream& in, GPoint& p) {
  in >> p.x >> p.y;
  return in;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GRect::set(float x_, float y_, float width_, float height_) {
  x = x_; y = y_; w = width_; h = height_;
}

void GRect::translate(float dx, float dy) {
  x += dx;
  y += dy;
}

GRect::GRect(GPoint const& p1, GPoint const& p2) {
  if (p1.x <= p2.x) {x = p1.x; w = p2.x - p1.x + 1;}
  else {x = p2.x; w = p1.x - p2.x + 1;}

  if (p1.y <= p2.y) {y = p1.y; h = p2.y - p1.y + 1;}
  else {y = p2.y; h = p1.y - p2.y + 1;}
}

bool GRect::contain(GPoint const& p, float tol) const {
  return
  p.x >= x-tol && p.x <= x+w+tol
  && p.y >= y-tol && p.y <= y+h+tol;
}

bool GRect::intersect(GPoint const& p1, GPoint const& p2) const {
  return
  linesIntersect(GPoint(x,y), GPoint(x+w,y), p1, p2)
  || linesIntersect(GPoint(x,y), GPoint(x,y+h), p1, p2)
  || linesIntersect(GPoint(x+w,y), GPoint(x+w,y+h), p1, p2)
  || linesIntersect(GPoint(x,y+h), GPoint(x+w,y+h), p1, p2);
}

bool GRect::operator==(const GRect& other) const {
  return x == other.x && y == other.y && w == other.w && h == other.h;
}

bool GRect::operator!=(const GRect& other) const {
  return x != other.x || y != other.y || w != other.w || h != other.h;
}

std::ostream& operator<<(std::ostream& out, const GRect& p) {
  out << p.x << " "<< p.y << " "<< p.w << " "<< p.h;
  return out;
}

std::istream& operator>>(std::istream& in, GRect& p) {
  in >> p.x >> p.y >> p.w >> p.h;
  return in;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GPoints::translate(float dx, float dy) {
  for (auto& it : *this)  {
    it.x += dx;
    it.y += dy;
  }
}

GRect GPoints::frame() const {
  float xmin{99999}, ymin{99999};
  float xmax{0}, ymax{0};
  for (auto& it : *this)  {
    xmax = std::max(xmax, it.x);
    ymax = std::max(ymax, it.y);
    xmin = std::min(xmin, it.x);
    ymin = std::min(ymin, it.y);
  }
  return GRect(xmin, ymin, xmax-xmin, ymax-ymin);
}

// treats as closed
bool GPoints::contain(GPoint const& p) const {
  return (evaluateCrossings(*this, p.x, p.y, false, BIG_VALUE) & 1) != 0;
}

/*
 bool GPoints::contains(float x, float y) const {
 return (evaluateCrossings(*this, x, y, false, BIG_VALUE) & 1) != 0;
 }*/

/* Test if a high-precision rectangle lies completely in the shape. This is
 * true if all points in the rectangle are in the shape. This implementation
 * is precise.
 */
bool GPoints::contain(GRect const& r) const {
  //if (!closed || !getBounds().intersects(r)) return false;
  // Does any edge intersect?
  if (evaluateCrossings(*this, r.x, r.y, false, r.w) != 0 /* top */
      || evaluateCrossings(*this,r.x, r.y + r.h, false, r.w) != 0 /* bottom */
      || evaluateCrossings(*this,r.x + r.w, r.y, true, r.h) != 0 /* right */
      || evaluateCrossings(*this,r.x, r.y, true, r.h) != 0) /* left */
    return false;
  // No intersections, is any point inside?
  if ((evaluateCrossings(*this, r.x, r.y, false, BIG_VALUE) & 1) != 0) return true;
  return false;
}

bool GPoints::intersect(const GRect& r) const {
  // does any edge intersect?
  if (evaluateCrossings(*this,r.x, r.y, false, r.w) != 0 /* top */
      || evaluateCrossings(*this,r.x, r.y + r.h, false, r.w) != 0 /* bottom */
      || evaluateCrossings(*this,r.x + r.w, r.y, true, r.h) != 0 /* right */
      || evaluateCrossings(*this,r.x, r.y, true, r.h) != 0) /* left */
    return true;
  // no intersections, is any point inside?
  if ((evaluateCrossings(*this,r.x, r.y, false, BIG_VALUE) & 1) != 0) return true;
  return false;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static int relativeCCW(GPoint p1, GPoint p2, GPoint p) {
  p2.x -= p1.x;
  p2.y -= p1.y;
  p.x -= p1.x;
  p.y -= p1.y;
  float ccw = p.x * p2.y - p.y * p2.x;
  
  if (ccw == 0.0) {
    // The point is colinear, classify based on which side of
    // the segment the point falls on. We can calculate a
    // relative value using the projection of PX,PY onto the
    // segment - a negative value indicates the point projects
    // outside of the segment in the direction of the particular
    // endpoint used as the origin for the projection.
    ccw = p.x * p2.x + p.y * p2.y;
    
    if (ccw > 0.0) {
      // Reverse the projection to be relative to the original p2.x,p2.y
      // p2.x and p2.y are simply negated.
      // PX and PY need to have (p2.x - p1.x) or (p2.y - p1.y) subtracted
      // from them (based on the original values)
      // Since we really want to get a positive answer when the
      // point is "beyond (p2.x,p2.y)", then we want to calculate
      // the inverse anyway - thus we leave p2.x & p2.y negated.
      p.x -= p2.x;
      p.y -= p2.y;
      ccw = p.x * p2.x + p.y * p2.y;
      if (ccw < 0.0) ccw = 0.0;
    }
  }
  return (ccw < 0.0) ? -1 : ((ccw > 0.0) ? 1 : 0);
}

static bool linesIntersect(const GPoint& a, const GPoint& b,
                           const GPoint& c, const GPoint& d) {
  return ((relativeCCW(a, b, c) * relativeCCW(a, b, d) <= 0)
          && (relativeCCW(c, d, a) * relativeCCW(c, d, b) <= 0));
}

/* calculates the number of intersections
 * between the polygon and a line extending from the point (x, y) along
 * the positive X, or Y axis, within a given interval.
 * returns the winding number.
 */
static int evaluateCrossings(GPoints const& points,
                             float x, float y, bool useYaxis, float distance) {
  // Get a value which is small but not insignificant relative the path.
  float epsilon = 1e-7f;
  GPoint p0, p1;
  int crossings = 0;
  
  if (useYaxis) {
    p0.x = points[0].y - y;
    p0.y = points[0].x - x;
  }
  else {
    p0.x = points[0].x - x;
    p0.y = points[0].y - y;
  }
  if (p0.y == 0.0) p0.y -= epsilon;
  
  for (size_t k = 2; k < points.size(); ++k) {
    if (useYaxis) {
      p1.x = points[k].y - y;
      p1.y = points[k].x - x;
    }
    else {
      p1.x = points[k].x - x;
      p1.y = points[k].y - y;
    }
    if (p1.y == 0.0) p1.y -= epsilon;
    if (p0.y * p1.y < 0) {
      if (linesIntersect(p0, p1, GPoint(epsilon,0.0), GPoint(distance,0.0))) {
        ++crossings;
      }
    }
    p0 = p1;
  }
  
  // end segment
  if (useYaxis) {
    p1.x = points[0].y - y;
    p1.y = points[0].x - x;
  }
  else {
    p1.x = points[0].x - x;
    p1.y = points[0].y - y;
  }
  if (p1.y == 0.0) p1.y -= epsilon;
  
  if (p0.y * p1.y < 0) {
    if (linesIntersect(p0, p1, GPoint(epsilon,0.0), GPoint(distance,0.0))) {
      ++crossings;
    }
  }
  return crossings;
}

}
