//
//  Base of GObjects (useful for srcipts among other things)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_ObjectBase_hpp
#define Guit_ObjectBase_hpp
#include <map>
#include <gobject.hpp>
#include <gptr.hpp>
namespace guit {

class GObjectBase {
public:
  enum Status{InvalidName = -1, NameExists = 0, OK = 1};

  GObjectBase();
  ~GObjectBase();

  /// Loads most GadgetType(s) - must be called before reading a Script.
  static void loadGadgetClasses();

  /// Adds a Gadget or GProp to the base.
  /// - The _object_ passed as an argument must NOT be destroyed.
  /// - The _name_ must be valid (standard C++ variable name) and no other object in the base
  ///  should have the same name, otherwise warning message and the object is not added
  /// @{
  Status add(GString const& name, GObject&);
  /// @}

  /// Adds a function to the base.
  /// - The _function_ passed as an argument is copied and can safely be destroyed.
  /// - The _name_ must be valid (standard C++ variable name) and no other object in the base
  ///  should have the same name, otherwise warning message and the function is not added
  Status add(GString const& name, GFunction const& function);
  
  /// Returns the object corresponding to this name (null if not found or improper type).
  /// @{
  GObject* getObject(GString const& name);
  Gadget*  getGadget(GString const& name);
  GBox*    getBox(GString const& name);
  GProp*   getProp(GString const& name);
  /// @}

  void printObjects(std::ostream&);

  /// @internal.
  /// @{
  using Map = std::map<GString,gptr<GObject>>;
  Map map_;
  /// @}
};


/// @internal.
class GFunctionWrapper : public GObject {
public:
  GFunctionWrapper(GFunction const& fun) : fun(fun) {}
  virtual GString typeName() const override;
  GFunction fun;
};

}
#endif
