//
// Hotkeys
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gevent.hpp>
#include <ghotkey.hpp>
#include <gwindow.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP_TYPE(Hotkey, GPropType::Hotkey)

GHotkey::GHotkey(GString const& value) {
  mapcode_ = 0;
  if (value.empty()) return;
  text_ = value;

  auto pos = value.rfind('-');
  if (pos == GString::npos) {
    value_.modifiers = GEventState::NoState;
    value_.key = ::tolower(value[0]);
  }
  else if (pos < value.size() -1) {
    value_.modifiers = GEvent::modifiersToStates(value.substr(0, pos));
    value_.key = ::tolower(value[pos+1]);
  }
}

GHotkey* GHotkey::clone(bool copy_value) const {
  auto* p = new GHotkey("");
  if (copy_value) {
    p->mapcode_ = mapcode_;
    p->value_ = value_;
  }
  return p;
}

void GHotkey::onRemove(Gadget* g) {
  if (g) g->settings_.hotkeySet = 0;
}

void GHotkey::apply(Gadget* g, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Hotkey]) return;
  r.specif[GPropType::Hotkey] = specif;

  r.hotkey_ = &text_;

  if (g && !g->settings_.hotkeySet) {
    //cerr << "GHotkey::apply "<< text_ << endl;
    g->settings_.hotkeySet = 1;
    // window devrait etre dans render
    if (auto* w = g->getWindow()) w->addHotkey(*g, *this);
  }
}

}
