//
//  Int Prop
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitInt_hpp
#define GuitInt_hpp
#include <gnumber.hpp>
namespace guit {

/// Integer prop.
class GInt : public GNumber_<int> {
public:
  using Super = GNumber_<int> ;

  GInt();
  GInt(int value);
  GInt(int value, int min, int max = highest());
  GInt(GInt const& value);
  GInt(GString const& value);  ///< null if _value_ is not a valid integral number.
  GInt* clone(bool copy_value) const override;

  GInt* toInt() override {return this;}

  /// Future value in an Active Expression (see GExpr).
  auto operator~() {return GNumPropRef_<GInt>(*this);}

  /// Binds this prop to an active expression (see GExpr).
  GInt& operator<<=(GNumExpr&);

  /// Specifies range.
  GInt& setRange(int min, int max) {Super::setRange(min,max); return *this;}
  GInt& setRangeF(float min, float max) override {Super::setRangeF(min,max); return *this;}

  /// Specifies whether the int cycles between min() and max()
  /// If _cycles_ is true, operator++ sets the value to min() if it becomes
  /// >= to max() (same thing the other way with operator--).
  GInt& setCycles(bool);

  GInt& operator=(int value) {set(value,false); return *this;}
  GInt& operator=(GInt const& value) {set(value,false); return *this;}
  GInt& operator=(GString const& value) {set(value); return *this;}

  GInt& operator++()    {increase(); return *this;}
  GInt& operator++(int) {increase(); return *this;}
  GInt& operator--()    {decrease(); return *this;}
  GInt& operator--(int) {decrease(); return *this;}
  GInt& operator+=(int delta) {increase(delta); return *this;}
  GInt& operator-=(int delta) {increase(-delta); return *this;}

  void increase(int const& delta) override;
  void increase() override {increase(+1);}
  void decrease() override {increase(-1);}

  using GType = GPropType_<GInt,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}
};


/// Return a GInt created in the heap.
/// @{
GInt& Int();
inline GInt& Int(int value) {return *new GInt(value);}
GInt& Int(int value, int min, int max = GInt::highest());
GInt& Int(GString const& value);
/// @}

}
#endif
