//
//  Charts
//  Guit GUI Toolkit
//  Copyright © 2019/2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <iostream>
#include <guit.hpp>
#include <gshape.hpp>
#include <gchart.hpp>
using namespace guit;
using namespace std;

class GUI {
  // value of the selected item and its range
  GFloat selvalue{0,0,1};

  // color of the selected item
  GBackground selcolor;

  // number of items
  GInt itemcount{1000};

  GScrollBox scrollbox{"#scrollbox"};
  GShow show_attributes{true}, show_value;

  GadgetType* lasttype{};

public:
  GUI() {
    auto& toolbar = ToolBar("#toolbar")
    << ("{icon:open.png}"
       // << (OpenDialog("Open File",":") << [&]{scrollbox.readValues(on::text());})
        )
    << ("{icon:save.png}"
       // << (SaveDialog("Save File","~") << [&]{scrollbox.writeValues(on::text());})
        )
    << ("{icon:print.jpg}"
       // << [&]{std::cout << scrollbox;}
        )
    << Filler
    << (ChoiceItem
        << on::checked   / "<b> < " / [this]{show_attributes = false;}
        << on::unchecked / "<b> > " / [this]{show_attributes = true;}
        );


    auto& createbox = ChoiceBox("{flow:row}")
    << ("Chart"       << [&]{createItems(GChartItem::Type());})
    << ("Rects"       << [&]{createItems(GRectShape::Type());})
    << ("Ovals"       << [&]{createItems(GOvalShape::Type());})
    << ("Buttons"     << [&]{createItems(GChoiceItem::Type());})
    << ("Radios"      << [&]{createItems(GRadioItem::Type());})
    << ("TextFields"  << [&]{createItems(GTextField::Type());});


    //
    auto& size_entry = TextField("{size: 6c a}");
    size_entry
    << itemcount
    << on::change / [&]{size_entry << GColor::orange;}
    << on::action / [&]{
      if (lasttype) createItems(*lasttype);
      size_entry << GColor::black;
    };


    auto& attrbox = VBox("#attrbox")
    << show_attributes
    << (VBox("#color")
        << makeSlider("<color=red><b>R", selcolor.redField())
        << makeSlider("<color=green><b>G", selcolor.greenField())
        << makeSlider("<color=blue><b>B", selcolor.blueField())
        << makeSlider("<b>A", selcolor.alphaField())
        << (Label("{size: 10c 20; border: rounded 5 5}") << (Background() <<= selcolor))
        )
    << (VBox("#value {margin:0}")
        << show_value
        << makeSlider("V", selvalue)
        );


    auto& controls = StatusBar("#typebox")
    << Label("Create:")
    << createbox
    << Label("Count:")
    << size_entry;


   auto& mainwin = Window("#gui")
    << toolbar
    << (HBox << scrollbox << attrbox)
    << controls;

    mainwin.startApp();
  }


  // !!! BUG ETRANGE: si N = 100 ca s'affiche pas dans GRID !!!

  void createItems(GadgetType& childtype) {
    lasttype = &childtype;
    scrollbox.removeChildren();
    GChoiceBox* box{};

    if (&childtype == &GChartItem::Type()) {
      show_value = true;
      auto* chart = new GChartBox("#chart {flow:row; flex:0 1;}");
      chart->setRange(0, 1);
      box = chart;
      for (int k = 0; k < itemcount; ++k) {
        *box << (ChartItem(float(k) / itemcount)
                 // << Background(float(k) / itemcount, 0, float(itemcount-k) / itemcount)
                 );
      }
    }
    else {
      show_value = false;
      box = new GChoiceBox("{flow:grid 100; flex:0 0}");
      if (dynamic_cast<GShape::GType*>(&childtype)) {
        for (int k = 0; k < itemcount; ++k) {
          *box << childtype.newInstance(".shape ");
        }
      }
      else {
        for (int k = 0; k < itemcount; ++k) {
          *box << childtype.newInstance("Item " + toString(k));
        }
      }
    }

    
    // note: a on:change callback cannot be used because we would get
    // the color of the item once checked, instead of the normal color.
    // Thus, we catch events and detect which item is selected on mousedown

    // NB empeche textfield de marcher
    box->catchEvents(true);
    *box << on::mousedown / [=](GEvent& e){downCB(*box, e.chosen());};

    scrollbox << box;
  }


  void downCB(GChoiceBox& box, Gadget* item) {
    if (!item) return;
    //cerr << "index: " << box.childIndex(item) << endl;

    // unbind selcolor with the color of the gadget that was previously selected
    selcolor.unbind(1);

    // item->prop<GBackground>(true) retreives the corresponding prop, taking
    // CSS into account (because of the true argument). This prop is added
    // to the gadget, or, if the gadget already has a prop with this type,
    // the prop's value is changed accordingly
    gsync(selcolor, item->prop<GBackground>(true), 1);

    if (show_value()) {
      selvalue.unbind(1);
      if (auto* it = item->to<GChartItem>()) {
        gsync(selvalue, item->prop<GFloat>(), 1);
      }
    }

    box.choose(item);
  }

  Gadget& makeSlider(GString const& name, GFloat& value) {
    return HBox(".textslider")
    << Label(name)
    << (Slider << value)
    << (TextField << value);
  }

  Gadget& makeSlider(GString const& name, GColor::Field const& field) {
    // colors are in range [0, 1]
    auto& value = *new GFloat(0.,0.,1.);
    // sync value with field
    value ^= field;
    return makeSlider(name, value);
  }
};


int main(int argc, const char **argv) {
  GApp app(argc, argv);
  app <<
  "#gui {title:Perf; icon:icon.png; size:700 500; flow:row; margin:0 0}"
  
  "#attrbox {flex: 0 1; margin: 0 0; background:white}"
  ".panel {flex:1 0; border:rounded 5 5; align:s x; gap:5 5;"
  "        margin:3 0; background:darkgray}"
  "#value {:.panel; title:{text: Value; color:white; font: bold arial 18}}"
  "#color {:.panel; title:{text: Color; color:white; font: bold arial 18}}"
  
  "#attrbox HBox {margin: 0}"
  "#attrbox Slider {size: 80 a; flex: 0 0}"
  "#attrbox TextField {size: 4c a; flex: 0 0; border:rounded}"
  
  "#typebox {align:s c; gap:5; font: 14 arial}"
  "#scrollbox {background:white}"
  
  ".shape {size:30 20}"
  "#chart ChartItem {background:steelblue; size:10 10}"
  //"#chart ChartItem { size:10 10}"
  ;
  GUI gui;
  return app.status();
}
